/****************************************************************
 * mkdistributable.cc, February 2022
 *
 * This program cannot create blocksampler betas.  It can only produce
 * the running bswa.  The delta argument is 3.0 by default or the first
 * argument.  The decay argument is 2.0 or the second argument.
 *
 * It expects a **simple** CSV type file with variable names in the
 * first row:
 *
 * permno needs to be ordered in increasing sequence.
 * yyyymmdd needs to be ordered in increasing sequence
 * ret that are not numeric are interpreted as missing and silently ignored.
 * vwretd and rf need to be series, too.
 *
 * if there is a ticker, it will be passed through.
 ****************************************************************/

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <math.h>
}

#include "fregr.hh"

/****************************************************************
 * the key required function for a bsw beta
 ****************************************************************/
static inline double winsrel( double delta, double ri, double rm ) {
  double b1= (1.0+delta)*rm;
  double b2= (1.0-delta)*rm;
  if (b2 < b1) { double b3= b2; b2= b1; b1= b3; }
  if (ri < b1) return b1;
  if (ri > b2) return b2;
  return ri;
}


/****************************************************************
 * Gregorian Date Handling
 *
 * Note: A stock that has many days but ceases to trade for a year can
 * continue on and produce a market-beta with even one day after its
 * interim break data...because it had observations (very
 * downweighted) from long ago, it does not suffer the 3*(5-day WEEK)
 * minimum.
 *
 ****************************************************************/

int gregdt(const int yyyymmdd) {
  int y= yyyymmdd / 10000;
  int m= (yyyymmdd / 100) % 100;
  int d= yyyymmdd % 100;

  y+= 8000;
  if(m<3) { y--; m+= 12; }
  return (y*365) +(y/4) -(y/100) +(y/400) -1200820
    +(m*153+3)/5-92
    +d-1 - (2424881 -1 - 153) ; // 19251231 is 0
}


/****************************************************************/

#define LINELEN 300 // the maximum length of a line
#define SEP ','

int main(int argc, char *argv[]) {
  double delta= (argc<=1) ? 3.0 : atof(argv[1]);
  assert( delta > 1.0 );  // or else the 1-delta definition is silly
  double agedecay= (argc<=2) ? 2.0 : atof(argv[2]);

  fprintf(stderr, "\n%s: Freely Distributable.\n", argv[0]);
  fprintf(stderr, "\tSee also Ivo Welch, \"Simpler Better Market Betas,\" CFR 2022\n");
  fprintf(stderr, "\tParameters: delta=%g agedecay=%g\n\n", agedecay, delta);

  char line[LINELEN];

  fprintf(stderr, "[reading **SIMPLE** csv-type file from stdin]\n");
  FILE *fcrsp= stdin;
  assert( fcrsp != NULL );

  //---------------- parse the CSV style header line

  assert( fgets(line, LINELEN, fcrsp) != NULL );

  static int ipermno= (-99), irf=(-99), ivwretd=(-99), iret=(-99), iyyyymmdd=(-99), iticker=(-99);
  if (ipermno < 0) { 
    char *token = line-1; //strchr( line, SEP ); // warning: strtok cannot parse empty fields.  really.
    for (int fieldcount=0; token!=NULL; ++fieldcount) {
#define matches(x) ((strncmp(token+1, x, strlen(x))==0))
      if ((iticker < 0) && (matches("ticker"))) iticker= fieldcount;
      if ((iticker < 0) && (matches("tic"))) iticker= fieldcount; // allow also 'tick'
      if ((ipermno < 0) && (matches("permno"))) ipermno= fieldcount;
      if ((iyyyymmdd < 0) && (matches("yyyymmdd"))) iyyyymmdd= fieldcount;
      if ((irf < 0) && (matches("rf"))) irf= fieldcount;
      if ((ivwretd < 0) && (matches("vwretd"))) ivwretd= fieldcount;
      else if ((iret < 0) && (matches("ret"))) iret= fieldcount; 
#undef matches

      token = strchr(token+1, SEP);
    }
    fprintf(stderr, "Detected CSV Fieldnames: permno=%d yyyymmdd=%d rf=%d vwretd=%d ret=%d\n", ipermno, iyyyymmdd, irf, ivwretd, iret);
    assert( (ipermno>=0) && (irf>=0) && (ivwretd>=0) && (iret>=0) && (iyyyymmdd>=0) );
  }

  //---------------- ready for actual regression
  
  FREGR *freg= fregr_new( 1 );
  int linecount= 0; int numbetaout= 0;
  int prev_permno= 10000, prev_yyyymm= 198601, prev_yyyymmdd= 19860108, prev_gdi= 0; char prev_ticker[20];
  double curweight= 1.0;

  if (iticker >= 0) printf("tic,");
  printf("permno, yyyymmdd ,     n , bswa%g%g\n", delta, agedecay);

  int skipped=0, used=0;
  while (!feof(fcrsp)) {
    if (((++linecount) % 1000000) == 0) fprintf(stderr, "[%d:%d: num of betas sofar=%d]\n", prev_permno, prev_yyyymmdd, numbetaout);

    if (fgets(line, LINELEN, fcrsp) == NULL) break;
    if (feof(fcrsp)) break;

    int fieldctr=0;
    char *fieldpointers[100];

    {
      char *token = line-1; // strchr( line, SEP );
      for (fieldctr=0; token!=NULL; ++fieldctr) {
	fieldpointers[ fieldctr ]= token;
	token = strchr(token+1, SEP);
      }
    }

    char firstchar_on_ret= fieldpointers[ iret ][1];
    if ((firstchar_on_ret != '-') && (firstchar_on_ret != '+') && (!isdigit(firstchar_on_ret))) {
      // could be 'C', 'B' on CRSP, or empty
      // fprintf(stderr, "W: not a number, skipping '%c' %s\n", firstchar_on_ret, line);
      ++skipped;
      continue;
    } else { ++used; }

    int permno= atoi( fieldpointers[ ipermno ]+1 );
    int yyyymmdd= atoi( fieldpointers[ iyyyymmdd ]+1 );

    // pray that atof gets a valid number.  it returns 0.0 if not.
    float ret= atof( fieldpointers[ iret ]+1 );

    float rf= atof( fieldpointers[ irf ]+1 );
    float vwretd= atof( fieldpointers[ ivwretd ]+1 );
    
    double xvwretd= vwretd-rf; double xret= ret-rf;

    assert( (yyyymmdd > 19200000) && (yyyymmdd < 22000000) );
    assert( xret > (-2.0) );

    assert( permno >= prev_permno );

    int gdi= gregdt( yyyymmdd );
    const int DAYSPERYEAR= 365;

    if (permno == prev_permno ) assert( gdi > prev_gdi );

    int yyyymm= yyyymmdd/100;

    // look for at least one week worth of data!!  warning: obscure cases
    const int WEEK=5; // deays
    if (((yyyymm != prev_yyyymm) || (permno != prev_permno))   &&   (freg->n > (3*WEEK))) {
      // trailing day beta to be printed
      double beta= fregr_regress(freg)[2];
      beta= winsrel( delta, beta, 1.0 );  // some pathological cases --- do not allow betas to be off by more than delta

      if (iticker >= 0) printf("%s,", prev_ticker);

      printf("%d , %8d , %5d , %7.3f\n", prev_permno, prev_yyyymmdd, freg->n, beta);  // could add fregr->n
      ++numbetaout;
    }

    if (permno != prev_permno) {
      fregr_clear(freg);
      curweight= 1.0;
    }

    prev_permno= permno; 
    prev_yyyymmdd= yyyymmdd;
    prev_yyyymm= yyyymm;
    if (iticker >= 0) {
#define save_to_sep( d, s ) { int i; char *vs= s; for (i=0; (vs[i]!=SEP) & (vs[i]!='\0'); ++i) d[i]=vs[i]; d[i++]='\0'; }
      save_to_sep( prev_ticker, fieldpointers[ iticker ] + 1);
    }

    // a few weird cases with intermissions means we cannot simply multiply by 1+agedecay/365
    // curweight*= pow( (1.0+agedecay/365), gdi-prev_gdi );

    for (int i=0; i<(gdi-prev_gdi); ++i) curweight*= (1.0+agedecay/DAYSPERYEAR);
    prev_gdi= gdi;


    // everything before was book-keeping; now comes the actual regression input

    double sxret= (delta < 999.9) ? winsrel( delta, xret, xvwretd ) : xret;
    double rm[]= { 1.0, xvwretd };

    fregr_include( freg, curweight, sxret, rm );

#ifdef DEBUG
    if ((permno == 93436)) {  // 93426 is tesla
      static FILE *FDEBUGFIRM= NULL;
      if (FDEBUGFIRM==NULL) { FDEBUGFIRM= fopen("tsla-debug.csv", "w"); fprintf(FDEBUGFIRM, "yyyymmdd , curweight, vwretd   , ret       , sxret\n"); }
      if (yyyymmdd <= 20150000) // do not disclose too much CRSP data
	fprintf(FDEBUGFIRM, "%d , %lf , %lf , %lf , %lf\n", yyyymmdd, curweight, vwretd, xret, sxret );
    }
#endif

  }

  // repeat
  // trailing day beta to be printed
  double beta= fregr_regress(freg)[2];
  beta= winsrel( delta, beta, 1.0 );  // some pathological cases --- do not allow betas to be off by more than delta

  if (iticker >= 0) printf("%s,", prev_ticker);
  printf("%d , %8d , %5d , %7.3f\n", prev_permno, prev_yyyymmdd, freg->n, beta);  // could add fregr->n
  ++numbetaout;

  fprintf(stderr, "DONE with #delta=%.2f #agedecay=%.2f --> num betas=%d (%d rets skipped, %d used)\n", delta, agedecay, numbetaout, skipped, used);
  fprintf(stderr, "suggest mkdistributable-post.R now\n");
  return 0;
}
